use crate::api;
use crate::core;
use crate::repositories;
use diesel::{
  r2d2::{ConnectionManager, Pool},
  PgConnection,
};

///
pub struct Synchronizer {
  pool: Pool<ConnectionManager<PgConnection>>,
  api_config: core::config::ApiConfig,
  synchronization_history_repository: repositories::synchronization_history::repository::Repository,
  vendor_repository: repositories::vendor::repository::Repository,
  vendor_api: api::vendor::api::Api,
}

impl std::fmt::Debug for Synchronizer {
  fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
    f.debug_struct("Synchronizer of api.kgon.identity").finish()
  }
}

impl std::clone::Clone for Synchronizer {
  fn clone(&self) -> Self {
    Self {
      pool: self.pool.clone(),
      api_config: self.api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      vendor_repository: repositories::vendor::repository::Repository::new(),
      vendor_api: api::vendor::api::Api::new(self.api_config.clone()),
    }
  }
}

impl Synchronizer {
  ///
  pub fn new(
    pool: Pool<ConnectionManager<PgConnection>>,
    api_config: core::config::ApiConfig,
  ) -> Synchronizer {
    Synchronizer {
      pool,
      api_config: api_config.clone(),
      synchronization_history_repository:
        repositories::synchronization_history::repository::Repository::new(),
      vendor_repository: repositories::vendor::repository::Repository::new(),
      vendor_api: api::vendor::api::Api::new(api_config),
    }
  }

  pub async fn vendors(&self) -> Result<(), Box<dyn std::error::Error>> {
    let start_at = (chrono::Utc::now()).timestamp();

    if let Err(e) = async {
      let conn = self.pool.get().expect("conn");

      let req = api::vendor::models::ListVendorsRequest {};
      let res = self.vendor_api.list_vendors(req).await?;

      let upsert_vendors: Vec<repositories::vendor::models::UpsertVendor> = res
        .vendors
        .iter()
        .map(|d| repositories::vendor::models::UpsertVendor {
          id: d.id,
          company_id: d.company_id,
          vendor_id: d.vendor_id,
          key: d.key.clone(),
          name: d.name.clone(),
          category: d.category.clone(),
          max_bet_casino: d.max_bet_casino,
          max_bet_slot: d.max_bet_slot,
          is_enable: d.is_enable.clone(),
          bet_count: d.bet_count,
        })
        .collect();

      let _affected = self
        .vendor_repository
        .upserts(&conn, upsert_vendors)
        .expect("vendor upsert");

      self
        .synchronization_history_repository
        .insert(
          &conn,
          &repositories::synchronization_history::models::NewSynchronizationHistory {
            item: repositories::synchronization::models::ITEM_VENDORS.to_string(),
            start_at,
            complete_at: (chrono::Utc::now()).timestamp(),
            code: 0,
            message: None,
          },
        )
        .expect("synchronization_history insert");

      Ok::<(), api::core::models::Error>(())
    }
    .await
    {
      let conn = self.pool.get().expect("conn");
      self
        .synchronization_history_repository
        .insert(
          &conn,
          &repositories::synchronization_history::models::NewSynchronizationHistory {
            item: repositories::synchronization::models::ITEM_VENDORS.to_string(),
            start_at,
            complete_at: (chrono::Utc::now()).timestamp(),
            code: e.code,
            message: e.msg,
          },
        )
        .expect("synchronization_history insert");
    }

    Ok(())
  }
}
